#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include <stdio.h>
#include <dlfcn.h>

#include "PDF417Pro_Types.h"
#include "LoadBMP.c"

static void Decode(char* imageFilePath, PPDF417_Options pdec_with_options);

///////////////////////////////////////////////////////////////////////////////////
//
// Description of pointers to functions called from the library
//
typedef PPDF417_Decoder (*PtrConnect_PDF417_Decoder)(int, int );
typedef void          (*PtrDisconnect_PDF417_Decoder)(PPDF417_Decoder&);
typedef PPDF417_Options (*PtrCreate_PDF417_Options)  (PPDF417_Decoder, TPDF417_OptMode);
typedef void          (*PtrDelete_PDF417_Options)    (PPDF417_Options&);
typedef int           (*PtrDecode_PDF417_Bits)         (PPDF417_Options ,int ,int ,TRow*);
typedef TPDF417_ImageInfo* (*PtrGetPDF417_ImageInfo)  (PPDF417_Options);
typedef TPDF417_Info*  (*PtrGetPDF417_Info)           (PPDF417_Options ,int);

PtrConnect_PDF417_Decoder    pConnect_PDF417_Decoder = NULL;
PtrDisconnect_PDF417_Decoder pDisconnect_PDF417_Decoder = NULL;
PtrCreate_PDF417_Options     pCreate_PDF417_Options = NULL;
PtrDelete_PDF417_Options     pDelete_PDF417_Options = NULL;
PtrDecode_PDF417_Bits        pDecode_PDF417_Bits = NULL;
PtrGetPDF417_ImageInfo       pGetPDF417_ImageInfo = NULL;
PtrGetPDF417_Info            pGetPDF417_Info = NULL;
//////////////////////////////////////////////////////////////////////////////////////

bool LoadLibrary(void)
{
  void* handle = dlopen("libPDF417.so", RTLD_LAZY);
  if(handle == NULL)
  {
   printf("Failed dlopen\n");
   return false;
  }
  if((pConnect_PDF417_Decoder    = (PtrConnect_PDF417_Decoder)   dlsym(handle, "Connect_PDF417_Decoder")) == NULL) return false;
  if((pDisconnect_PDF417_Decoder = (PtrDisconnect_PDF417_Decoder)dlsym(handle, "Disconnect_PDF417_Decoder")) == NULL) return false;
  if((pCreate_PDF417_Options     = (PtrCreate_PDF417_Options)    dlsym(handle, "Create_PDF417_Options")) == NULL) return false;
  if((pDelete_PDF417_Options     = (PtrDelete_PDF417_Options)    dlsym(handle, "Delete_PDF417_Options")) == NULL) return false;
  if((pDecode_PDF417_Bits        = (PtrDecode_PDF417_Bits)        dlsym(handle, "Decode_PDF417_Bits")) == NULL) return false;
  if((pGetPDF417_ImageInfo       = (PtrGetPDF417_ImageInfo)      dlsym(handle, "GetPDF417_ImageInfo")) == NULL) return false;
  if((pGetPDF417_Info            = (PtrGetPDF417_Info)           dlsym(handle, "GetPDF417_Info")) == NULL) return false;
  return true;
}

// the images size
const int maxrowQ = 4000;       // maximum of image rows
const int maxcolQ = 4000;

static TRow pbits[maxrowQ]; // image lines
static TRow pmembits; // buffer for image


int main(int argc, char* argv[])
{
   PPDF417_Decoder     pdecoder;
   PPDF417_Options     pdec_with_options1;
   PPDF417_Options     pdec_with_options2;
   TPDF417_OptMode     optmode;

   if(argc != 2)
   {
    printf("Usage:\n");
    printf("demo.out MyImageName.bmp\n");
    return 0;
   }
   
   if(!LoadLibrary())
   {
     printf("Failed LoadLibrary!\n");
     return 0;
   }
   
   pmembits = (TRow)malloc(maxrowQ * maxcolQ);

   for (int row = 0; row < maxrowQ; row++)
   {
      pbits[row] = &(pmembits[maxcolQ*row]);
   }
   // Construct decoder:
   pdecoder = pConnect_PDF417_Decoder(maxrowQ, maxcolQ);

   optmode.MaxSymbolCount   = 1;
   //optmode.SpeedMode    = SP_Robust;
   optmode.SpeedMode    = SP_Fast;
   optmode.CalcGrade    = false;
   optmode.DirectMode   = DM_LeftRight;
   optmode.BWFilter     = 0;
   optmode.ScanStep     = 4;
   optmode.TimeOut      = 0;

   // Construct the options:
   pdec_with_options1 = pCreate_PDF417_Options(pdecoder, optmode);

   optmode.BWFilter     = -1; // bar wudth reduction
   pdec_with_options2 = pCreate_PDF417_Options(pdecoder, optmode);

   // Decoding
   printf("======== Option1 ========\n");
   Decode(argv[1], pdec_with_options1);

   printf("\n======= Option2 =========\n");
   Decode(argv[1], pdec_with_options2);

   pDelete_PDF417_Options(pdec_with_options1);
   pDelete_PDF417_Options(pdec_with_options2);
   pDisconnect_PDF417_Decoder (pdecoder);

   free(pmembits);
}

void Decode(char* imageFilePath, PPDF417_Options pdec_with_options)
{
  int rowcount = maxrowQ;
  int colcount = maxcolQ;
  TPDF417_ImageInfo*  pimageinfo;
  TPDF417_Info*       pinfo;
  clock_t beg, end;

  if(LoadBMP(imageFilePath, pbits, &rowcount, &colcount) != 0)
  {
    printf("Failed load bmp file: %s\n", imageFilePath);
    return;
  }
  // Decode
  beg = clock();
  pDecode_PDF417_Bits(pdec_with_options, rowcount, colcount, pbits);
  end = clock();

  // Get decode results:
  pimageinfo = pGetPDF417_ImageInfo(pdec_with_options);

  if(pimageinfo->RejectionReason == PDF417_OK)
  {
    printf("PDF417 count = %4d\n", pimageinfo->SymbolCount);
    for(int labelId = 0; labelId < pimageinfo->SymbolCount; labelId++)
    {
      pinfo = pGetPDF417_Info(pdec_with_options, labelId);
      // display pinfo ...
      printf("Decoded array: %s\n", pinfo->pch);
    }
  }
  else
  {
    printf("RejectionReason = %d\n", pimageinfo->RejectionReason);
  }
  printf("Time = %d\n", (int)((end - beg)*1000 / CLOCKS_PER_SEC));
}